from .borders import Bdr
from .card import Card
from .fill import Fill
from .flex import Flex
from .hr import HR
from .line import Line
from .pretty import pretty
from .pretty_exception import PrettyException
from .printer import Printer
from .table import Table, CellDfn
from .text import Text
from . import context
print = Printer()



from threading import current_thread


def ctx_vars(reset=False):
    ''' These can't be tied to any particular object and must be consistant throughout the program
    '''
    try:
        assert(not reset)
        return current_thread()._ctx_vars_231174
    except:
        vars = {}
        current_thread()._ctx_vars_231174 = vars
        return vars



class CallableVar():
    pass



class ObjectAttr(CallableVar):
    def __init__(self, attr, default=None):
        self.attr = attr
        self.default = default

    def __call__(self, obj):
        try: return getattr(obj, self.attr)
        except AttributeError: return self.default



class CVar():
    ''' This is the prototype/class for a context variable.
    These get registered and must be uniuqe. i.e. 'style' can mean only
    one thing in the program
    '''
    def __init__(self, *names):
        self.names = tuple([names[0]] + sorted(names[1:]))


    def __hash__(self):
        return hash(self.names)

    
    def __eq__(self, other):
        return self.names == other.names


    def __repr__(self):
        return f"{self.__class__.__name__}({','.join(map(repr,self.names))})"


    def canon(self, sval):
        ''' Convert from *any* representation of the value to the canonical value.

        The value will never be ``None``.
        '''
        return sval



class IntCVar(CVar):
    def canon(self, sval):
        return int(sval) 


class FloatCVar(CVar):
    def canon(self, sval):
        return float(sval) 


class BoolCVar(CVar):
    def canon(self, sval):
        if not isinstance(sval, str): return bool(sval)
        return sval.lower().strip() not in ['', 'f','false','off','no','0']


class EnumCVar(CVar):
    def __init__(self, *args, options):
        super().__init__(*args)
        self.options = options

    def canon(self, sval):
        if sval not in self.options: raise ValueError(f"EnumCvar value {sval!r} not in {self.options!r}")
        return sval
    


from .text import Text
from .rich import Rich
from .span import Span
from .line import Line
from .size import Size
from .context import Context, ObjectAttr, CVar, BoolCVar, EnumCVar, FloatCVar


Context.define(BoolCVar('flex_reverse'))
Context.define(BoolCVar('flex_wrap'))
Context.define(EnumCVar('flex_dir', options=('-','|')))
Context.define(FloatCVar('height_rate', 'hs'))
Context.define(FloatCVar('width_rate', 'ws'))



class Flex(Rich):
    ''' A list of objects

    :Attributes:
        just:
            flex Justification:
                `<` : Left
                `.` : Center
                `>` : Right
                `#` : Space between
                `%` : Space around
        flex_reverse:
            Render the elements in reverse order

    :Examples:

        
    '''    
    def __init__(self, *args, **kwargs):
        self._cells = [[]]
        super().__init__(*args, **kwargs)


    @property
    def cells(self):
        if not self.rich_stream: return self._cells[:-1] if self._cells[-1] == [] else self._cells
        if self._cells[-1] != []: self._cells.append([self._cells.pop()])
        while self.rich_stream:
            elctx = self.rich_stream.pop(0)
            if elctx[0] == '\t':
                self._cells.append([])
            else:
                self._cells[-1].append(elctx)
        # Form arrays of (el,ctx) into a single Context object
        for i, els in enumerate(reversed(self._cells)):
            if i==0 and els==[]: continue # skip trailing []
            if isinstance(els, Context): break
            if els and isinstance(els[0], Context):
                self._cells[-1-i] = els.pop(0)
            elif els and isinstance(els[0][0], Context):
                el, ctx = els.pop(0)
                el = el.ctx_parent(self)
                el.ctx(**ctx)
                self._cells[-1-i] = el
            else:
                self._cells[-1-i] = Text(parent=self)
            for el, ctx in els:
                self._cells[-1-i](el, **ctx)
        return self._cells[:-1] if self._cells[-1] == [] else self._cells


    def clone(self, **kwargs):
        s = self.__class__(**kwargs)
        s._cells = [c.clone(parent=s, **c.ctx_local) for c in self.cells]
        if self._cells[-1] == []: s._cells.append([])
        return s


    def each_child(self):
        yield from self.cells


    def _flatten_width_row(self, cells, w, h):
        flat = [list(c.user.flatten(w=c.size,h=h)) for c in cells]
        if h == 0:
            mh = len(flat[0]) if len(flat) == 1 else max(*[len(x) for x in flat])
            flat = [f if len(f) == mh else list(c.user.flatten(w=c.size,h=mh)) for f, c in zip(flat, cells)]
        return [Line(*f, parent=self).justify(w) for f in zip(*flat)]


    def _flatten_width(self, cells, wrap, w, h):
        def elide_cols(cells):
            dots = '~' if self['ascii'] else '…'
            if w > 30: dots = f'{dots}{len(cells)}{dots}'
            s = Line(style="dem", parent=self).insert(0, dots)
            return Size(nom=s.width, rate=0, user=s)

        def elide_rows(cells):
            vbar = '|' if self['ascii'] else '⋮'
            s = Line(parent=self, style="dem", justify='|').insert(0, f'{vbar}{len(cells)}{vbar}')
            if w and s.width > w: s = Line(parent=self, style="dem", justify='|').insert(0, vbar)
            return Size(nom=1, rate=0, user=[Size(size=w, nom=s.width, user=s)])

        rows = Size.resize(cells, w, wrap, elide_cols)
        flats = [self._flatten_width_row(row, w, 0) for row in rows]
        collapse = [z for y in flats for z in y]
        if not h or len(collapse) == h: return collapse
        # Our height doesn't fit, so perform another resize for the cross axis
        hcells = [Size(nom=len(flat), user=row) for row,flat in zip(rows,flats)]
        col = Size.resize(hcells, h, wrap=False, elide=elide_rows)[0]
        flats = [self._flatten_width_row(c.user, w, c.size) for c in col]
        collapse = [z for y in flats for z in y]
        assert(len(collapse) == h)
        return collapse


    _keys = ('nom', 'min', 'max', 'rate')

    def flatten(self, w=0, h=0, **kwargs):
        wrap = self['flex_wrap']
        els = list(reversed(self.cells) if self['flex_reverse'] else self.cells)
        if not els: return []
        d = 'width' if (self['flex_dir'] or '-') == '-' else 'height'
        def _size(e):
            args = {k:e.ctx(f'{d}_{k}') for k in Flex._keys}
            return Size(user=e, **{k:v for k,v in args.items() if v != None})
        cells = [_size(e) for e in els]
        if d == 'width': return self._flatten_width(cells, wrap, w, h)
        raise NotImplementedError()


    def tab(self, **kwargs):
        _ = self.cells
        if self._cells[-1] == []:self._cells.pop()
        self._cells.append(Text(parent=self, **kwargs))
        return self



import pytest
from print_ext.line import Line, Just, justify_v, SMark as SM, style_cvar
from print_ext.text import Text
from print_ext.span import Span
from print_ext.context import Context
from .testutil import debug_dump, styled


def Ln(*args, **kwargs):
    if isinstance(args[-1], int):
        kwargs['style'] = tuple(str(args[-1]))
        args=args[:-1]
    s = Line(**kwargs)
    for arg in reversed(args):
        s.insert(0, arg)
    return s


class EmptyStr():
    def __str__(self):
        return ''


class EmptyClass(Context):
    def __init__(self, val):
        self.val = [Line()] if val else []
        super().__init__()

    def flatten(self):
        return self.val


class StyleCtx(Context, style='x-y'):
    pass


def test_ctx_style():
    c1 = Context(style='a-b-c')
    c2 = StyleCtx(parent=c1)
    c3 = Context(parent=c2, style='b-c-d')
    print('\n'.join(debug_dump(c3)))
    assert(c3['style'] == style_cvar.canon('x-y-a-b-c-d'))



def test_StyleCVar():
    assert(style_cvar.canon('') == None)
    assert(style_cvar.canon('bob-') == ('',))
    assert(style_cvar.canon('-bob') == ('','bob'))
    assert(style_cvar.canon('-bob--') == ('',))
    assert(style_cvar.canon('--') == ('',))
    assert(style_cvar.canon('-') == ('',))
    assert(style_cvar.canon(('a','','b')) == ('','b'))
    assert(style_cvar.canon('bob--cob') == ('','cob'))
    with pytest.raises(ValueError):
        style_cvar.canon('bob cob')



def test_StyleCVar_merge():
    def _v(parent, val, result):
        assert(style_cvar.merge(style_cvar.canon(val), style_cvar.canon(parent)) == style_cvar.canon(result))
    _v('a-b', '-c-d', '-c-d')
    _v('a-b', 'b-c-d', 'a-b-c-d')
    _v('a-b', 'a-b', 'a-b')
    _v('a-b-a-b', 'a-b', 'a-b-a-b')
    _v('a-b-c', 'a-b', 'a-b-c-a-b')



def test_line_basic():
    s = Line('abc','カタカナ', 33)
    assert(str(s) == 'abcカタカナ33')
    


def test_empty_line():
    s = Line()
    assert(not s)
    assert(not s(''))
    assert(not s())
    assert(s('x'))



def testinsert_empty():
    s = Line()
    for v in ['\b3', '', EmptyClass(False), EmptyClass(True), EmptyStr()]:
        s.append(v)
        assert(repr(s) == 'Line()')



def test_line_clone():
    ''' When adding objects to a line they need to be reparented
    '''
    s1 = Ln('ab', 1)
    s1P = Ln(s1, 'cd', 2)
    assert(s1P.styled() == ('abcd', [SM('2',0,4), SM('1',0,2)]))
    s2 = Ln('ef', 3)
    s2P = Ln(s2, 'gh', 4)
    s = Ln(s1, s2, 5)
    s2P['style'] = '6'
    print('\n'.join(debug_dump(s)), '\n\n-----------\n')
    assert(s.styled() == ('abef', [SM('5',0,4), SM('2',0,2), SM('1',0,2), SM('4',2,4), SM('3',2,4)]))
    assert(s2P.styled() == ('efgh', [SM('6',0,4), SM('3',0,2)]))



def test_line_flatten_clone():
    ''' When a line is flattened it should still be parented correctly
    '''
    s = Line('abc', style='bob')
    rows = list(s.flatten())
    assert(rows[0].parent == s)



def test_line_append_multiline():
    s = Line(Text('bob'))
    assert(str(s) == 'bob')
    assert(str(Line(Text('bob\vcob'))) == 'bob\\ncob')
    assert(str(Line("the\vquick\vbrown")) == 'the\\nquick\\nbrown')



def test_line_append_tab():
    assert(str(Line('the\tquick')) == 'the    quick')



def test_justify_h():
    ''' Line.justify()
    '''
    s = Line('bob', justify=':')
    s.justify(4)
    assert(str(s) == ' bob')
    s.justify(4,justify_h='<')
    assert(str(s) == ' bob')
    s.justify(7,justify_h='|')
    assert(str(s) == '  bob  ')
    with pytest.raises(ValueError):
        s.justify(2)
    


def test_line_no_wrap():
    def _f(s, **kwargs):
        return styled(Line(s, justify='|', ascii='yes', text_wrap=False), **kwargs)[0]
    assert(_f('abcdefghij', w=13) == (' abcdefghij  ', []))
    assert(_f('abcdefghij', w=5) == ('ab~ij', [SM('dem',2,3)]))
    assert(_f('abcdefghij', w=4) == ('ab~j', [SM('dem',2,3)]))
    assert(_f('abcdefghij', w=3) == ('a~j',[SM('dem',1,2)]))
    assert(_f('abcdefghij', w=2) == ('a~',[SM('dem',1,2)]))
    assert(_f('abcdefghij', w=1) == ('~',[SM('dem',0,1)]))
    assert(_f('abcdefghij', w=0) == ('abcdefghij', []))
    assert(_f('aあa', w=4) == ('aあa', []))
    assert(_f('ああa', w=3) == ('~~a', [SM('dem',0,2)]))
    assert(_f('ああa', w=2) == ('~~', [SM('dem',0,2)]))
    assert(_f('test', w=3) == ('t~t', [SM('dem',1,2)]))



def test_line_loop():
    l = Line('hi',style='1')
    l(l,style='2')
    assert(l.styled() == ('hihi',[SM('1',0,4), SM('2',2,4), SM('1',2,4)]))


def test_line_loop_deep():
    b = Line('b', style='1')
    abc = Line('a', b, 'c', style='2')
    assert(abc.styled() == ('abc', [SM('2',0,3), SM('1',1,2)]))
    b(abc)
    print('\n'.join(debug_dump(b)))
    assert(b.styled() == ('babc', [SM('2',0,4), SM('1',0,4), SM('2',1,4), SM('1',2,3)]))


def test_just_defaults():
    j = Just()
    assert((j.h, j.v) == ('',''))
    j = Just(defaults='|-')
    assert((j.h, j.v) == ('|','-'))
    j = Just('_', defaults='|-')
    assert((j.h, j.v) == ('|','_'))
    j = Just('<', defaults='|-')
    assert((j.h, j.v) == ('<','-'))


def test_just_pad():
    j = Just('|-')
    assert(j.pad_v(10) == 5)
    assert(j.pad_h(10) == 5)
    assert(j.pad_v(9) == 4)
    assert(j.pad_h(9) == 4)



def test_line_wrap():
    def _f(s, w=0, h=0, just='|-', wmf=5, **kwargs):
        return [r.styled() for r in  Line(s, justify=just, ascii='y', text_wrap=True, wrap_mark_from=wmf, **kwargs).flatten(w, h)]
    assert(_f('abcdefghij', w=13) == [(' abcdefghij  ',[])])
    assert(_f('abcdefghij', w=3) == [('abc',[]), ('def',[]), ('ghi',[]), ('j  ',[])])
    assert(_f('abcdefghij', w=6) == [('abcdef',[]), ('\\ ghij',[SM('dem',0,2)])])
    assert(_f('abcdefghijabcdefgh', w=11, h=1) == [(' |2 lines| ',[SM('dem',0,11)])])
    assert(_f('abcdefghijabcdefghijkl', w=11, h=2) == [('abcdefghija',[]),(' |2 lines| ',[SM('dem',0,11)])])
    assert(_f('012345678012345601234560123456', w=9, h=3) == [('012345678',[]),('|2 lines|',[SM('dem',0,9)]),('\\ 0123456',[SM('dem',0,2)])])
    assert(_f('1111222233334444555566667777888899990000aaaa', w=4, h=1) == [('|11 ',[SM('dem',0,4)])])
    assert(_f('1111', w=1, h=1) == [('|',[SM('dem',0,1)])])
    #assert(_f('00000111112222233333', w=5, h=2, lang='ja', wmf=10) == [('00000',[]),('|3行|',[SM('dem',0,4)])])



def test_line_split():
    def s(*parts): return Line(*parts)
    def f(i, *parts): return str(s(*parts).trim(i))
    assert(f(-3, "abc", "あ",s('12型34')) == '34')
    assert(f(0, "abc", "あ",s('12型34')) == '')
    assert(f(3, "abc", "あ",s('12型34')) == 'abc')
    assert(f(5, "abc", "あ",s('12型34')) == 'abcあ')
    assert(f(-6, "abc", "あ",s('12型34')) == '12型34')
    assert(f(-7, "abc", "あ",s('12型34')) == '12型34')
    assert(f(6, "abc", "あ",s('12型34')) == 'abcあ1')



def test_line_trim_styles():
    ''' trimming keeps styles
    '''
    s = Ln('abc', Ln('def',1), 'ghi', 2)
    print('\n'.join(debug_dump(s)))
    assert(s.styled() == ('abcdefghi', [SM('2',0,9), SM('1',3,6)]))
    s.trim(4)
    assert(s.styled() == ('abcd', [SM('2',0,4), SM('1',3,4)]))
    s.trim(3)
    assert(s.styled() == ('abc', [SM('2',0,3)]))



def test_line_trim():
    def _f(x): return repr(Ln(Ln('abc'), 'def', Ln('ghi')).trim(x))
    assert(_f(1) == "Line(Line(Span('a')))")
    assert(_f(1000) == "Line(Line(Span('abc')), Span('def'), Line(Span('ghi')))")
    assert(_f(0) == "Line()")
    assert(_f(4) == "Line(Line(Span('abc')), Span('d'))")
    assert(_f(-4) == "Line(Span('f'), Line(Span('ghi')))")
    assert(_f(-1) == "Line(Line(Span('i')))")




def test_style_flatten():
    p = Line('test parent', style='p')
    s = Line('abc', parent=p)
    assert(s.styled() == ('abc', [SM('p',0,3)]))
    s.append('yz')
    assert(s.styled() == ('abcyz', [SM('p',0,5)]))
    s.insert(0, '  ')
    assert(s.styled() == ('  abcyz', [SM('p',0,7)]))
    s.insert(0, Line("o", style='1-2'))
    print('_'*80)
    print('\n'.join(debug_dump(s)), s.width)
    assert(s.styled() == ('o  abcyz', [SM('p',0,8), SM('1',0,1), SM('2',0,1)]))



def test_styled_double():
    s = Ln(Ln('a',1), Ln('b',1))
    assert(s.styled() == ('ab', [SM('1',0,2)]))



def test_extend_two():
    s = Ln(Ln(Ln('a',1),Ln('b',2),3), Ln('c',32), 6)
    #print('\n'.join(debug_dump(s)))
    assert(s.styled() == ('abc', [SM('6',0,3), SM('3',0,3), SM('1',0,1), SM('2',1,3)]))



def test_diffently_parented():
    p1 = Context(style='1')
    p2 = Context(style='2')
    s1 = Line(parent=p1, style='9').insert(0, 'a')
    s2 = Line(parent=p2).insert(0, 'b')
    s3 = Ln(s1,s2,3)
    assert(s3.styled() == ('ab', [SM('3',0,2), SM('1',0,1), SM('9',0,1), SM('2',1,2)]))



def test_pop_one_match_two():
    s = Ln(Ln('a',1), Ln('b',2), Ln('c',31), 3)
    assert(s.styled() == ('abc', [SM('3',0,3), SM('1',0,1), SM('2',1,2), SM('1',2,3)]))



def test_over_extend():
    s = Ln(Ln(Ln('a',1), Ln('b',2), 3), Ln('c',2))
    assert(s.styled() == ('abc', [SM('3',0,2), SM('1',0,1), SM('2',1,2), SM('2',2,3)]))



def test_pop_extend():
    ''' after popping  1  but we are appending 13 so just append 3'''
    s = Ln(Ln('a',2), Ln('b',13), 1)
    assert(s.styled() == ('ab', [SM('1',0,2), SM('2',0,1), SM('3',1,2)]))



def test_pop_extend_nested():
    ''' after popping  121  but we are appending 12123 so just append 23'''
    s = Ln(Ln('a',121), Ln('b',12123))
    assert(s.styled() == ('ab', [SM('1',0,2), SM('2',0,2), SM('1',0,2), SM('2',1,2),SM('3',1,2)]))



def test_pop_before_extend():
    ''' 23236 pop 236  new 236  -> 236? or 23236?'''
    s = Ln(Ln('a',2), Ln('b',523),5)
    assert(s.styled() == ('ab', [SM('5',0,2), SM('2',0,2), SM('3',1,2)]))



def test_pop_extend_new():
    ''' 123 pop 3 new 234   2 is parent 3 is extend 4 is new '''
    s = Ln(Ln('a',3), Ln('b',234),12)
    assert(s.styled() == ('ab', [SM('1',0,2), SM('2',0,2), SM('3',0,2), SM('4',1,2)]))



def test_extend_partial():
    ''' 123 pop 23 new 12.  only extend 2 '''
    s = Ln(Ln('a',23), Ln('b',12), 1)
    assert(s.styled() == ('ab', [SM('1',0,2), SM('2',0,2), SM('3',0,1)]))



def test_styled_nested():
    ''' bug: nested spans dont get styled
    '''
    s = Line('bob', style='p').insert(0, 'abc')
    assert(s.styled() == ('abcbob', [SM('p',0,6)]))
    s.insert(0, Line('xyz', style='1-2'))
    assert(s.styled() == ('xyzabcbob', [SM('p',0,9), SM('1',0,3), SM('2',0,3)]))



def test_styled_dedup_sibling():
    s = Line('\b0$ abc', 'def')
    assert(s.styled() == ('abcdef', [SM('0',0,6)]))



def test_style_justify():
    ''' justified space should not be included in the style '''
    s = Line('abc', style='s', justify='|')
    assert(styled(s, w=7) == [('  abc  ', [SM('s',0,7)])])



def test_subtyles():
    s1 = Line('abc', style='a')
    s2 = Line('def', style='b')
    s3 = Line(s1, s2, style='x-y')
    assert(s3.styled() == ('abcdef', [SM('x',0,6),SM('y',0,6),SM('a',0,3),SM('b',3,6)]))



def test_justify_v():
    rows = [Line('a'), Line('b'), Line('c'), Line('d')]
    r = list(justify_v(rows, 4, Just('-'), Line('xxxx')))
    assert('--'.join(map(str,r)) == 'a--b--c--d')
    r = list(justify_v(rows, 6, Just('-'), Line('xxx')))
    assert('--'.join(map(str,r)) == 'xxx--a--b--c--d--xxx')
    with pytest.raises(ValueError):
        list(justify_v(rows, 3, '-',Line()))
    assert(r[0].styled() == ('xxx', []))
    assert(r[1].styled() == ('a', []))
    assert(r[5].styled() == ('xxx', []))


if __name__=='__main__':
    test_extend_two()


import pytest
from print_ext.flex import Flex
from print_ext.fill import Fill
from print_ext.text import Text
from print_ext.line import Line, SMark as SM
from .testutil import debug_dump, styled

def f(o, w,h):
    return ','.join(map(str, o.flatten(w=w, h=h)))


def test_flex_clone():
    f = Flex('a\tb\vc\td')
    print('\n'.join(debug_dump(f)))
    print('-----\n', *[f'-{s.styled()[0]}-\n' for s in f.flatten()])
    print('-------\n', f.width, f.height, f.ctx('height_nom', 'width_nom'))
    f2 = Flex(f, '\t', f)
    print('\n\n\n\n')
    print('\n'.join(debug_dump(f2)))
    lines = [str(x) for x in f2.flatten()]
    assert(lines == ['abdabd', ' c  c '])



def test_flex_in_flex():
    p = Flex('bob\vcob\tfoo\vbar')
    f = Flex('ab\tcd\t',p)
    lines = [str(x) for x in f.flatten()]
    assert(lines == ['abcdbobfoo', '    cobbar'])


def test_flex_loops():
    g = Flex('hi')
    i = Flex(g,g)
    lines = [str(x) for x in i.flatten()]
    assert(lines == ['hihi'])



def test_flex_fox():
    fox = Flex(style='1', text_wrap=True, wrap_mark_from=10, justify='>_').tab(wx=5, style='2')('\b3 the',' quick\vbrown\t', 'fox','\t',Fill('y'))
    els = list(fox.each_child())
    print(f"{len(els)} Flex cells")
    for e in els:
        print(f'   {type(e)} {e}')
    el0 = list(els[0].each_child())
    print('\n'.join(debug_dump(els[0])))
    assert(el0[0]['width_max'] == 5)
    assert(el0[0]['width_nom'] == 9)
    assert(els[2]['width_nom'] == 0)
    assert(f(fox, 20, 0) == 'the q         yyyyyy,uick          yyyyyy,brown      foxyyyyyy')
    


def test_flex_small():
    ''' The flex itsself is justified
    '''
    fox = Flex(Fill('x',width_max=3), '\t', Text('y\vz',width_max=4, justify='>'), justify='|')
    assert(f(fox, 10, 0) == ' xxx   y  , xxx   z  ')



@pytest.mark.xfail(reason='not implemented')
def test_flex_large():
    ''' The flex is too large
    '''
    fox = Flex(Fill('x', width_min=3), '\t', Fill('a\nb', width_min=3))
    assert(f(fox, 2, 0) == 'xxxxxaaaaa,xxxxxbbbbb')
   


def test_flex_wrap():
    def fox(**kwargs): return Flex(Fill('123456789'), '\t', Fill('a\vb', width_min=3), '\t', Fill('z', width_min=3), **kwargs)
    assert(f(fox(), 10, 0) == '1234aaazzz,1234bbbzzz')
    assert(f(fox(), 10, 2) == '1234aaazzz,1234bbbzzz')
    assert(f(fox(flex_wrap=True), 10, 3) == '1234aaazzz,1234bbbzzz,1234aaazzz')



def test_flex_height():
    def fox(**kwargs): return Flex(Fill('123456789', width_min=9), '\t', Fill('a\vb', width_min=3), '\t', Fill('z', width_min=3), **kwargs)
    assert(f(fox(flex_wrap=True, ascii='y'), 10, 2) == '1234567891,   |1|    ')
    assert(f(fox(flex_wrap=True), 10, 5) == '1234567891,1234567891,aaaaazzzzz,bbbbbzzzzz,aaaaazzzzz')
    assert(f(Flex('aaaaa\tbbbbb\tccccc\tddddd\teeeee', ascii=1, flex_wrap=True), 5, 3) == 'aaaaa, |3| ,eeeee')


def test_flex_elide():
    assert(f(Flex('a\tb\tc\td\te\tf', ascii='y'), 3, 0) == 'a~f')


def test_flex_str():
    f = Flex('\berr ERROR',' j/k')
    print('\n'.join(debug_dump(f)))
    print('!'*80)
    fl = f.flatten()[0]
    print('\n'.join(debug_dump(fl)))
    print('-'*80)
    assert(fl.styled() == ('ERROR j/k', [SM('err',0,5)]))
    

def test_flex_empty():
    f = Flex()
    assert((f.width, f.height) == (0,0))
    f.flatten()


def test_flex_passthrough():
    l = Line('hi')
    f = Flex(l)
    assert(styled(l,w=10) == styled(f,w=10))



from .printer import Printer
from .text import Text
from .table import Table
from .pretty import pretty


class PrettyException(Exception):
    ''' This implements the pretty() method to show a pretty version of the exception.
    
    __str__ and __repr__ return normal, not-pretty, strings.
    '''
    def __init__(self, **kwargs):
        for k,v in kwargs.items(): setattr(self, k, v)


    def __str__(self):
        for line in Printer(ascii=True, color=False).each_line(pretty(self)):
            return str(line)
        else:
            return repr(self)


    def __repr__(self):
        s = f"{self.__class__.__name__}("
        args = [f'{k}={v!r}' for k,v in self.__dict__.items()]
        return s+ ', '.join(args) + ')'


    def pretty(self):
        t = Text(self.__class__.__name__)
        if self.__dict__:
            t('\v',pretty(self.__dict__))
        return t


